from typing import Dict, Union

from flask import Blueprint, request
from sqlalchemy import asc, desc, func

from fittrackee import appLog
from fittrackee.oauth2.server import require_auth
from fittrackee.responses import (
    HttpResponse,
    InvalidPayloadErrorResponse,
    NotFoundErrorResponse,
    UserNotFoundErrorResponse,
)

from .exceptions import (
    FollowRequestAlreadyProcessedError,
    NotExistingFollowRequestError,
)
from .models import FollowRequest, User

follow_requests_blueprint = Blueprint("follow_requests", __name__)

FOLLOW_REQUESTS_PER_PAGE = 10
MAX_FOLLOW_REQUESTS_PER_PAGE = 50


@follow_requests_blueprint.route("/follow-requests", methods=["GET"])
@require_auth(scopes=["follow:read"])
def get_follow_requests(auth_user: User) -> Dict:
    """
    Get follow requests to process, received by authenticated user.

    **Scope**: ``follow:read``

    **Example requests**:

    - without parameters

    .. sourcecode:: http

      GET /api/follow-requests/ HTTP/1.1

    - with some query parameters

    .. sourcecode:: http

      GET /api/follow-requests?page=1&order=desc  HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "data": {
          "follow_requests": [
            {
              "admin": false,
              "bio": null,
              "birth_date": null,
              "created_at": "Thu, 02 Dec 2021 17:50:48 GMT",
              "first_name": null,
              "followers": 1,
              "following": 1,
              "last_name": null,
              "location": null,
              "nb_sports": 0,
              "nb_workouts": 0,
              "picture": false,
              "records": [],
              "sports_list": [],
              "total_distance": 0.0,
              "total_duration": "0:00:00",
              "username": "Sam"
            }
          ]
        },
        "pagination": {
          "has_next": false,
          "has_prev": false,
          "page": 1,
          "pages": 1,
          "total": 1
        },
        "status": "success"
      }


    :query integer page: page if using pagination (default: 1)
    :query integer per_page: number of follow requests per page
                             (default: 10, max: 50)
    :query string order: sorting order (default: ``asc``)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 500:

    """
    params = request.args.copy()
    page = int(params.get("page", 1))
    per_page = int(params.get("per_page", FOLLOW_REQUESTS_PER_PAGE))
    order = params.get("order", "asc")
    if per_page > MAX_FOLLOW_REQUESTS_PER_PAGE:
        per_page = MAX_FOLLOW_REQUESTS_PER_PAGE
    follow_requests_pagination = (
        FollowRequest.query.filter_by(
            followed_user_id=auth_user.id,
            updated_at=None,
        )
        .order_by(
            asc(FollowRequest.created_at)
            if order == "asc"
            else desc(FollowRequest.created_at)
        )
        .paginate(page=page, per_page=per_page, error_out=False)
    )
    follow_requests = follow_requests_pagination.items
    return {
        "status": "success",
        "data": {
            "follow_requests": [
                follow_request.serialize()["from_user"]
                for follow_request in follow_requests
            ]
        },
        "pagination": {
            "has_next": follow_requests_pagination.has_next,
            "has_prev": follow_requests_pagination.has_prev,
            "page": follow_requests_pagination.page,
            "pages": follow_requests_pagination.pages,
            "total": follow_requests_pagination.total,
        },
    }


def process_follow_request(
    auth_user: User, user_name: str, action: str
) -> Union[Dict, HttpResponse]:
    from_user = User.query.filter(
        func.lower(User.username) == func.lower(user_name),
    ).first()
    if not from_user:
        appLog.error(
            f"Error when accepting follow request: {user_name} does not exist"
        )
        return UserNotFoundErrorResponse()

    try:
        if action == "accept":
            auth_user.approves_follow_request_from(from_user)
        else:  # action == 'reject'
            auth_user.rejects_follow_request_from(from_user)
    except NotExistingFollowRequestError:
        return NotFoundErrorResponse(message="Follow request does not exist.")
    except FollowRequestAlreadyProcessedError:
        return InvalidPayloadErrorResponse(
            message=(
                f"Follow request from user '{user_name}' already {action}ed."
            )
        )

    return {
        "status": "success",
        "message": f"Follow request from user '{user_name}' is {action}ed.",
    }


@follow_requests_blueprint.route(
    "/follow-requests/<user_name>/accept", methods=["POST"]
)
@require_auth(scopes=["follow:write"])
def accept_follow_request(
    auth_user: User, user_name: str
) -> Union[Dict, HttpResponse]:
    """
    Accept a follow request from user.

    **Scope**: ``follow:write``

    **Example requests**:

    .. sourcecode:: http

      POST /api/follow-requests/Sam/accept HTTP/1.1

    **Example responses**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "status": "success",
        "message": "Follow request from user 'Sam' is accepted.",
      }

    :param string user_name: user name

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 400:
        - ``Follow request from user 'user_name' already accepted.``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user does not exist``
        - ``Follow request does not exist.``

    """
    return process_follow_request(auth_user, user_name, "accept")


@follow_requests_blueprint.route(
    "/follow-requests/<user_name>/reject", methods=["POST"]
)
@require_auth(scopes=["follow:write"])
def reject_follow_request(
    auth_user: User, user_name: str
) -> Union[Dict, HttpResponse]:
    """
    Reject a follow request from user.

    **Scope**: ``follow:write``

    **Example requests**:

    .. sourcecode:: http

      POST /api/follow-requests/Sam/reject HTTP/1.1

    **Example responses**:

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

      {
        "status": "success",
        "message": "Follow request from user 'Sam' is rejected.",
      }

    :param string user_name: user name

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: success
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 400:
        - ``Follow request from user 'user_name' already rejected.``
    :statuscode 403:
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``user does not exist``
        - ``Follow request does not exist.``

    """
    return process_follow_request(auth_user, user_name, "reject")



import os
from datetime import datetime, timezone
from logging import Logger
from typing import Optional

from dramatiq.middleware import Shutdown, TimeLimitExceeded
from dramatiq_abort import Abort
from humanize import naturalsize

from fittrackee import db, dramatiq
from fittrackee.constants import TASKS_TIME_LIMIT, TaskPriority
from fittrackee.exceptions import TaskException
from fittrackee.users.models import Notification, UserTask
from fittrackee.utils import decode_short_id
from fittrackee.workouts.services.workouts_from_file_creation_service import (
    WorkoutsFromArchiveCreationAsyncService,
)

GENERIC_ERROR = "error during archive processing"
ABORT_ERROR = "task execution aborted"


def update_task_and_clean(
    *,
    error: str,
    upload_task: Optional["UserTask"] = None,
    upload_task_id: Optional[int] = None,
) -> None:
    if not upload_task and not upload_task_id:
        return
    if upload_task is None:
        upload_task = UserTask.query.filter_by(id=upload_task_id).first()
    if upload_task is None:
        return

    if error == ABORT_ERROR:
        upload_task.aborted = True
    else:
        upload_task.errored = True
        upload_task.errors = {
            **upload_task.errors,
            "archive": error,
        }
    db.session.commit()

    if (
        not upload_task.aborted
        and not Notification.query.filter_by(
            event_object_id=upload_task.id,
            event_type=upload_task.task_type,
        ).first()
    ):
        notification = Notification(
            from_user_id=upload_task.user_id,
            to_user_id=upload_task.user_id,
            created_at=datetime.now(tz=timezone.utc),
            event_object_id=upload_task.id,
            event_type=upload_task.task_type,
        )
        db.session.add(notification)
        db.session.commit()

    if upload_task.file_path and os.path.exists(upload_task.file_path):
        os.remove(upload_task.file_path)


@dramatiq.actor(
    queue_name="fittrackee_workouts",
    priority=TaskPriority.MEDIUM,
    time_limit=TASKS_TIME_LIMIT,
    max_retries=0,
    notify_shutdown=True,
)
def upload_workouts_archive(task_id: int) -> None:
    try:
        service = WorkoutsFromArchiveCreationAsyncService(task_id)
        service.process()
    except (Abort, Shutdown, TimeLimitExceeded) as e:
        db.session.rollback()
        if isinstance(e, Abort):
            error = ABORT_ERROR
        elif isinstance(e, TimeLimitExceeded):
            error = "upload execution time exceeding limit"
        else:
            error = GENERIC_ERROR
        update_task_and_clean(error=error, upload_task_id=task_id)
        raise TaskException(error) from None
    except Exception as e:
        db.session.rollback()
        update_task_and_clean(error=GENERIC_ERROR, upload_task_id=task_id)
        raise TaskException(GENERIC_ERROR) from e
    finally:
        db.session.close()


def _handle_upload_task(upload_task: "UserTask", logger: "Logger") -> None:
    files_count = len(upload_task.data.get("files_to_process", []))
    file_size = str(upload_task.file_size) if upload_task.file_size else "0"
    logger.info(
        f"Processing task '{upload_task.short_id}' (files: {files_count}, "
        f"size: {naturalsize(file_size)})"
    )
    try:
        service = WorkoutsFromArchiveCreationAsyncService(upload_task.id)
        service.process()
        db.session.refresh(upload_task)
        archive_error = upload_task.errors.get("archive")
        if archive_error:
            logger.info(f" > archive error: {archive_error}")
        files_error = upload_task.errors.get("files", {})
        if files_error:
            logger.info(f" > errored files: {len(files_error.keys())}")
    except (KeyboardInterrupt, Exception) as e:
        db.session.rollback()
        error = (
            ABORT_ERROR if isinstance(e, KeyboardInterrupt) else GENERIC_ERROR
        )
        update_task_and_clean(error=error, upload_task_id=upload_task.id)
        raise e


def process_workouts_archives_uploads(max_count: int, logger: Logger) -> int:
    count = 0
    upload_tasks = (
        db.session.query(UserTask)
        .filter(
            UserTask.progress == 0,
            UserTask.errored == False,  # noqa
            UserTask.aborted == False,  # noqa
            UserTask.task_type == "workouts_archive_upload",
        )
        .order_by(UserTask.created_at)
        .limit(max_count)
        .all()
    )

    for upload_task in upload_tasks:
        _handle_upload_task(upload_task, logger)
        count += 1
        logger.info(" > done.")

    return count


def process_workouts_archive_upload(
    task_short_id: str, logger: Logger
) -> None:
    try:
        task_id = decode_short_id(task_short_id)
    except Exception as e:
        raise TaskException("Invalid task id") from e

    upload_task = UserTask.query.filter_by(
        uuid=task_id, task_type="workouts_archive_upload"
    ).first()
    if not upload_task:
        raise TaskException("No task found")
    if upload_task.progress > 0 or upload_task.errored or upload_task.aborted:
        raise TaskException("Task is not queued")

    _handle_upload_task(upload_task, logger)



from typing import TYPE_CHECKING, List

import requests

from .base_elevation_service import BaseElevationService
from .exceptions import ElevationServiceException

if TYPE_CHECKING:
    from gpxpy.gpx import GPXTrackPoint


MAX_POINTS = 15000


class OpenElevationService(BaseElevationService):
    """
    Documentation:
    https://github.com/Jorl17/open-elevation/blob/master/docs/api.md
    """

    config_key = "OPEN_ELEVATION_API_URL"
    url_pattern = "{base_url}/api/v1/lookup"
    log_label = "Open Elevation API"

    def _get_elevations_for_api(
        self, points: List["GPXTrackPoint"], smooth: bool = False
    ) -> List[int]:
        if not self.url:
            raise ElevationServiceException("Open Elevation API: no URL set")

        elevations: List[int] = []
        for i in range(0, len(points), MAX_POINTS):
            r = requests.post(
                self.url,
                json={
                    "locations": [
                        {
                            "latitude": point.latitude,
                            "longitude": point.longitude,
                        }
                        for point in points[i : i + MAX_POINTS]
                    ]
                },
                timeout=30,
            )
            r.raise_for_status()
            results = r.json().get("results", [])
            elevations = [*elevations, *[int(r["elevation"]) for r in results]]

        return elevations



"""add index on likes tables

Revision ID: d3c5338bedff
Revises: d3e40c2bda80
Create Date: 2025-01-04 17:17:55.637816

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'd3c5338bedff'
down_revision = 'd3e40c2bda80'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('comment_likes', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_comment_likes_comment_id'), ['comment_id'], unique=False)

    with op.batch_alter_table('workout_likes', schema=None) as batch_op:
        batch_op.create_index(batch_op.f('ix_workout_likes_workout_id'), ['workout_id'], unique=False)

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workout_likes', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_workout_likes_workout_id'))

    with op.batch_alter_table('comment_likes', schema=None) as batch_op:
        batch_op.drop_index(batch_op.f('ix_comment_likes_comment_id'))

    # ### end Alembic commands ###



"""update elevations_processing enum

Revision ID: ef18bfbcc613
Revises: c59425f77851
Create Date: 2025-11-12 12:15:42.977410

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ef18bfbcc613'
down_revision = 'c59425f77851'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.execute("ALTER TYPE elevations_processing RENAME TO elevations_processing_old")
    op.execute("CREATE TYPE elevations_processing AS ENUM('NONE', 'OPEN_ELEVATION', 'OPEN_ELEVATION_SMOOTH', 'VALHALLA')")
    op.execute(
        """
        ALTER TABLE users 
            ALTER COLUMN missing_elevations_processing DROP DEFAULT,
            ALTER COLUMN missing_elevations_processing TYPE elevations_processing USING missing_elevations_processing::text::elevations_processing,
            ALTER COLUMN missing_elevations_processing SET DEFAULT 'NONE';
        ALTER TABLE workouts 
            ALTER COLUMN missing_elevations_processing DROP DEFAULT,
            ALTER COLUMN missing_elevations_processing TYPE elevations_processing USING missing_elevations_processing::text::elevations_processing,
            ALTER COLUMN missing_elevations_processing SET DEFAULT 'NONE';
    """
    )
    op.execute("DROP TYPE elevations_processing_old")
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###

    op.execute("UPDATE users SET missing_elevations_processing = 'NONE' WHERE missing_elevations_processing = 'VALHALLA';")
    op.execute("UPDATE workouts SET missing_elevations_processing = 'NONE' WHERE missing_elevations_processing = 'VALHALLA';")
    op.execute(
        "ALTER TYPE elevations_processing RENAME TO elevations_processing_old")
    op.execute(
        "CREATE TYPE elevations_processing AS ENUM('NONE', 'OPEN_ELEVATION', 'OPEN_ELEVATION_SMOOTH')")
    op.execute(
        """
        ALTER TABLE users 
            ALTER COLUMN missing_elevations_processing DROP DEFAULT,
            ALTER COLUMN missing_elevations_processing TYPE elevations_processing USING missing_elevations_processing::text::elevations_processing,
            ALTER COLUMN missing_elevations_processing SET DEFAULT 'NONE';
        ALTER TABLE workouts 
            ALTER COLUMN missing_elevations_processing DROP DEFAULT,
            ALTER COLUMN missing_elevations_processing TYPE elevations_processing USING missing_elevations_processing::text::elevations_processing,
            ALTER COLUMN missing_elevations_processing SET DEFAULT 'NONE';
    """
    )
    op.execute("DROP TYPE elevations_processing_old")




"""add geometry and points to workout segments

Revision ID: fed271354410
Revises: b0ef6ca5ec92
Create Date: 2025-05-28 10:59:17.126175

"""
from alembic import op
import sqlalchemy as sa
from fittrackee.database import TZDateTime
from geoalchemy2 import Geometry

# revision identifiers, used by Alembic.
revision = 'fed271354410'
down_revision = '8221053f0e81'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workout_segments', schema=None) as batch_op:
        batch_op.add_geospatial_column(sa.Column('geom', Geometry(geometry_type='LINESTRING', srid=4326, dimension=2, spatial_index=False, from_text='ST_GeomFromEWKT', name='geometry'), nullable=True))
        batch_op.add_column(sa.Column('points', sa.JSON(), server_default='[]', nullable=False))
        batch_op.add_column(sa.Column('uuid', sa.UUID(), server_default=sa.text('gen_random_uuid ()'), nullable=False))
        batch_op.add_column(sa.Column('start_date', TZDateTime(), nullable=True))
        batch_op.create_geospatial_index('idx_workout_segments_geom', ['geom'], unique=False, postgresql_using='gist', postgresql_ops={})
        batch_op.create_index(batch_op.f('ix_workout_segments_start_date'), ['start_date'], unique=False)
        batch_op.create_unique_constraint('workout_segments_uuid_key', ['uuid'])
        batch_op.create_unique_constraint('workout_id_start_date_unique', ['workout_id', 'start_date'])

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('workout_segments', schema=None) as batch_op:
        batch_op.drop_constraint('workout_id_start_date_unique', type_='unique')
        batch_op.drop_constraint('workout_segments_uuid_key', type_='unique')
        batch_op.drop_index(batch_op.f('ix_workout_segments_start_date'))
        batch_op.drop_geospatial_index('idx_workout_segments_geom', postgresql_using='gist', column_name='geom')
        batch_op.drop_column('start_date')
        batch_op.drop_column('uuid')
        batch_op.drop_column('points')
        batch_op.drop_geospatial_column('geom')

    # ### end Alembic commands ###



"""update oauth2_code table

Revision ID: b30c7818c902
Revises: 67ae8866b18f
Create Date: 2025-09-27 13:21:54.189272

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b30c7818c902'
down_revision = '67ae8866b18f'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('oauth2_code', schema=None) as batch_op:
        batch_op.add_column(sa.Column('acr', sa.Text(), nullable=True))
        batch_op.add_column(sa.Column('amr', sa.Text(), nullable=True))

    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    with op.batch_alter_table('oauth2_code', schema=None) as batch_op:
        batch_op.drop_column('amr')
        batch_op.drop_column('acr')

    # ### end Alembic commands ###



"""add sport preferences

Revision ID: 080acc8ee956
Revises: 9842464bb885
Create Date: 2021-11-12 10:20:23.786727

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '080acc8ee956'
down_revision = '9842464bb885'
branch_labels = None
depends_on = None


def upgrade():

    op.drop_constraint('sports_img_key', 'sports', type_='unique')
    op.drop_column('sports', 'img')

    op.create_table(
        'users_sports_preferences',
        sa.Column('user_id', sa.Integer(), nullable=False),
        sa.Column('sport_id', sa.Integer(), nullable=False),
        sa.Column('color', sa.String(length=50), nullable=True),
        sa.Column('is_active', sa.Boolean(), nullable=False),
        sa.Column(
            'stopped_speed_threshold',
            sa.Float(),
            nullable=False,
        ),
        sa.ForeignKeyConstraint(
            ['sport_id'],
            ['sports.id'],
        ),
        sa.ForeignKeyConstraint(
            ['user_id'],
            ['users.id'],
        ),
        sa.PrimaryKeyConstraint('user_id', 'sport_id'),
    )


def downgrade():
    op.drop_table('users_sports_preferences')

    op.add_column(
        'sports',
        sa.Column(
            'img', sa.VARCHAR(length=255), autoincrement=False, nullable=True
        ),
    )
    op.create_unique_constraint('sports_img_key', 'sports', ['img'])
    op.execute(
        """
        UPDATE sports AS s
        SET img = si.img
        FROM (VALUES
            ('Cycling (Sport)','/img/sports/cycling-sport.png'),
            ('Cycling (Transport)','/img/sports/cycling-transport.png'),
            ('Hiking','/img/sports/hiking.png'),
            ('Mountain Biking','/img/sports/mountain-biking.png'),
            ('Running','/img/sports/running.png'),
            ('Walking','/img/sports/walking.png'),
            ('Mountain Biking (Electric)','/img/sports/electric-mountain-biking.png'),
            ('Trail','/img/sports/trail.png'),
            ('Skiing (Alpine)','/img/sports/alpine-skiing.png'),
            ('Skiing (Cross Country)','/img/sports/cross-country-skiing.png'),
            ('Rowing','/img/sports/rowing.png')
        ) AS si(label, img)
        WHERE si.label = s.label;
        """
    )



"""create Activity & Sport tables

Revision ID: b7cfe0c17708
Revises: 9741fc7834da
Create Date: 2018-01-21 17:24:52.587814

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'b7cfe0c17708'
down_revision = '9741fc7834da'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('sports',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('label', sa.String(length=50), nullable=False),
    sa.Column('img', sa.String(length=255), nullable=True),
    sa.Column('is_default', sa.Boolean(), default=False, nullable=False),
    sa.PrimaryKeyConstraint('id'),
    sa.UniqueConstraint('label')
    )
    op.create_table('activities',
    sa.Column('id', sa.Integer(), nullable=False),
    sa.Column('user_id', sa.Integer(), nullable=False),
    sa.Column('sport_id', sa.Integer(), nullable=False),
    sa.Column('title', sa.String(length=255), nullable=True),
    sa.Column('gpx', sa.String(length=255), nullable=True),
    sa.Column('creation_date', sa.DateTime(), nullable=True),
    sa.Column('modification_date', sa.DateTime(), nullable=True),
    sa.Column('activity_date', sa.DateTime(), nullable=False),
    sa.Column('duration', sa.Interval(), nullable=False),
    sa.Column('pauses', sa.Interval(), nullable=True),
    sa.Column('moving', sa.Interval(), nullable=True),
    sa.Column('distance', sa.Numeric(precision=6, scale=3), nullable=True),
    sa.Column('min_alt', sa.Numeric(precision=6, scale=2), nullable=True),
    sa.Column('max_alt', sa.Numeric(precision=6, scale=2), nullable=True),
    sa.Column('descent', sa.Numeric(precision=7, scale=2), nullable=True),
    sa.Column('ascent', sa.Numeric(precision=7, scale=2), nullable=True),
    sa.Column('max_speed', sa.Numeric(precision=6, scale=2), nullable=True),
    sa.Column('ave_speed', sa.Numeric(precision=6, scale=2), nullable=True),
    sa.ForeignKeyConstraint(['sport_id'], ['sports.id'], ),
    sa.ForeignKeyConstraint(['user_id'], ['users.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###

    op.execute(
        """
        INSERT INTO sports (label, img, is_default)
        VALUES 
            ('Cycling (Sport)', '/img/sports/cycling-sport.png', True),
            ('Cycling (Transport)', '/img/sports/cycling-transport.png', True),
            ('Hiking', '/img/sports/hiking.png', True),
            ('Mountain Biking', '/img/sports/mountain-biking.png', True),
            ('Running', '/img/sports/running.png', True),
            ('Walking', '/img/sports/walking.png', True)
        """
    )


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('activities')
    op.drop_table('sports')
    # ### end Alembic commands ###



"""Add 'notes' column to 'Activity' table

Revision ID: 096dd0b43beb
Revises: 71093ac9ca44
Create Date: 2018-06-13 17:56:20.359884

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '096dd0b43beb'
down_revision = '71093ac9ca44'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('activities', sa.Column('notes', sa.String(length=500), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('activities', 'notes')
    # ### end Alembic commands ###



import os
from datetime import datetime, timezone
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union
from uuid import UUID, uuid4

import jwt
from flask import current_app
from jsonschema import validate
from sqlalchemy import and_, func
from sqlalchemy.dialects import postgresql
from sqlalchemy.engine.base import Connection
from sqlalchemy.event import listens_for
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import Mapped, mapped_column, relationship
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.orm.session import Session, object_session
from sqlalchemy.schema import CheckConstraint
from sqlalchemy.sql import text
from sqlalchemy.sql.expression import select
from sqlalchemy.types import Enum

from fittrackee import BaseModel, appLog, bcrypt, db
from fittrackee.comments.models import Comment
from fittrackee.constants import ElevationDataSource, PaceSpeedDisplay
from fittrackee.database import TZDateTime
from fittrackee.dates import aware_utc_now
from fittrackee.files import get_absolute_file_path
from fittrackee.utils import encode_uuid
from fittrackee.visibility_levels import VisibilityLevel
from fittrackee.workouts.models import Workout
from fittrackee.workouts.utils.sports import get_sport_displayed_data

from .constants import (
    MESSAGE_PREFERENCES_SCHEMA,
    NOTIFICATION_TYPES,
    NOTIFICATIONS_PREFERENCES_SCHEMA,
    USER_LINK_TEMPLATE,
)
from .exceptions import (
    BlockUserException,
    FollowRequestAlreadyProcessedError,
    FollowRequestAlreadyRejectedError,
    InvalidNotificationTypeException,
    NotExistingFollowRequestError,
    UserTaskException,
    UserTaskForbiddenException,
)
from .roles import (
    UserRole,
    has_admin_rights,
    has_moderator_rights,
    is_auth_user,
)
from .utils.tokens import decode_user_token, get_user_token

if TYPE_CHECKING:
    from fittrackee.equipments.models import Equipment
    from fittrackee.reports.models import ReportAction
    from fittrackee.workouts.models import Record


TASK_TYPES = [
    "user_data_export",
    "workouts_archive_upload",
]

SEGMENTS_CREATION_EVENTS = [
    "all",
    "none",
    "only_manual",
]


class FollowRequest(BaseModel):
    """Follow request between two users"""

    __tablename__ = "follow_requests"
    follower_user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id"),
        primary_key=True,
    )
    followed_user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id"),
        primary_key=True,
    )
    is_approved: Mapped[bool] = mapped_column(default=False, nullable=False)
    created_at: Mapped[datetime] = mapped_column(
        TZDateTime, nullable=False, default=aware_utc_now
    )
    updated_at: Mapped[Optional[datetime]] = mapped_column(
        TZDateTime, nullable=True
    )

    to_user: Mapped["User"] = relationship(
        "User",
        primaryjoin="FollowRequest.followed_user_id == User.id",
        uselist=False,
        back_populates="received_follow_requests",
    )
    from_user: Mapped["User"] = relationship(
        "User",
        primaryjoin="FollowRequest.follower_user_id == User.id",
        uselist=False,
        back_populates="sent_follow_requests",
    )

    def __repr__(self) -> str:
        return (
            f"<FollowRequest from user '{self.follower_user_id}' "
            f"to user '{self.followed_user_id}'>"
        )

    def __init__(self, follower_user_id: int, followed_user_id: int):
        self.follower_user_id = follower_user_id
        self.followed_user_id = followed_user_id

    def is_rejected(self) -> bool:
        return not self.is_approved and self.updated_at is not None

    def serialize(self) -> Dict:
        return {
            "from_user": self.from_user.serialize(),
            "to_user": self.to_user.serialize(),
        }


@listens_for(FollowRequest, "after_insert")
def on_follow_request_insert(
    mapper: Mapper, connection: Connection, new_follow_request: FollowRequest
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Connection) -> None:
        to_user = User.query.filter_by(
            id=new_follow_request.followed_user_id
        ).first()
        if not to_user:
            return
        event_type = (
            "follow" if new_follow_request.is_approved else "follow_request"
        )

        if not to_user.is_notification_enabled(event_type):
            return

        notification = Notification(
            from_user_id=new_follow_request.follower_user_id,
            to_user_id=new_follow_request.followed_user_id,
            created_at=new_follow_request.created_at,
            event_type=event_type,
        )
        session.add(notification)


@listens_for(FollowRequest, "after_update")
def on_follow_request_update(
    mapper: Mapper, connection: Connection, follow_request: FollowRequest
) -> None:
    follow_request_object = object_session(follow_request)
    if follow_request_object and follow_request_object.is_modified(
        follow_request
    ):

        @listens_for(db.Session, "after_flush", once=True)
        def receive_after_flush(session: Session, context: Connection) -> None:
            if follow_request.is_approved:
                if follow_request.to_user.is_notification_enabled("follow"):
                    follow_request_notification = Notification.query.filter_by(
                        from_user_id=follow_request.follower_user_id,
                        to_user_id=follow_request.followed_user_id,
                        event_type="follow_request",
                    ).first()

                    if follow_request_notification:
                        notification_table = Notification.__table__  # type: ignore
                        connection.execute(
                            notification_table.update()
                            .where(
                                notification_table.c.from_user_id
                                == follow_request.follower_user_id,
                                notification_table.c.to_user_id
                                == follow_request.followed_user_id,
                                notification_table.c.event_type
                                == "follow_request",
                            )
                            .values(
                                event_type="follow",
                                marked_as_read=False,
                            )
                        )
                    else:
                        follow_notification = Notification(
                            from_user_id=follow_request.follower_user_id,
                            to_user_id=follow_request.followed_user_id,
                            created_at=datetime.now(timezone.utc),
                            event_type="follow",
                        )
                        session.add(follow_notification)

                if follow_request.from_user.is_notification_enabled(
                    "follow_request_approved"
                ):
                    notification = Notification(
                        from_user_id=follow_request.followed_user_id,
                        to_user_id=follow_request.follower_user_id,
                        created_at=datetime.now(timezone.utc),
                        event_type="follow_request_approved",
                    )
                    session.add(notification)

            if (
                not follow_request.is_approved
                and follow_request.updated_at is not None
            ):
                Notification.query.filter_by(
                    from_user_id=follow_request.follower_user_id,
                    to_user_id=follow_request.followed_user_id,
                    event_type="follow_request",
                ).delete()


@listens_for(FollowRequest, "after_delete")
def on_follow_request_delete(
    mapper: Mapper, connection: Connection, old_follow_request: FollowRequest
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        Notification.query.filter(
            Notification.from_user_id == old_follow_request.follower_user_id,
            Notification.to_user_id == old_follow_request.followed_user_id,
            Notification.event_type.in_(["follow", "follow_request"]),
        ).delete()
        Notification.query.filter(
            Notification.from_user_id == old_follow_request.followed_user_id,
            Notification.to_user_id == old_follow_request.follower_user_id,
            Notification.event_type == "follow_request_approved",
        ).delete()


class BlockedUser(BaseModel):
    __tablename__ = "blocked_users"
    __table_args__ = (
        db.UniqueConstraint(
            "user_id", "by_user_id", name="blocked_users_unique"
        ),
    )

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id", ondelete="CASCADE"),
        index=True,
    )
    by_user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id", ondelete="CASCADE"),
        index=True,
    )
    created_at: Mapped[datetime] = mapped_column(TZDateTime, nullable=False)

    def __init__(
        self,
        user_id: int,
        by_user_id: int,
        created_at: Optional[datetime] = None,
    ):
        self.user_id = user_id
        self.by_user_id = by_user_id
        self.created_at = (
            datetime.now(timezone.utc) if created_at is None else created_at
        )


class User(BaseModel):
    __tablename__ = "users"
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    username: Mapped[str] = mapped_column(
        db.String(255), unique=True, nullable=False
    )
    email: Mapped[str] = mapped_column(
        db.String(255), unique=True, nullable=False
    )
    password: Mapped[str] = mapped_column(db.String(255), nullable=False)
    created_at: Mapped[datetime] = mapped_column(TZDateTime, nullable=False)
    first_name: Mapped[Optional[str]] = mapped_column(
        db.String(80), nullable=True
    )
    last_name: Mapped[Optional[str]] = mapped_column(
        db.String(80), nullable=True
    )
    birth_date: Mapped[Optional[datetime]] = mapped_column(
        TZDateTime, nullable=True
    )
    location: Mapped[Optional[str]] = mapped_column(
        db.String(80), nullable=True
    )
    bio: Mapped[Optional[str]] = mapped_column(db.String(200), nullable=True)
    picture: Mapped[Optional[str]] = mapped_column(
        db.String(255), nullable=True
    )
    timezone: Mapped[Optional[str]] = mapped_column(
        db.String(50), nullable=True
    )
    date_format: Mapped[Optional[str]] = mapped_column(
        db.String(50), nullable=True
    )
    # weekm: does the week start Monday?
    weekm: Mapped[bool] = mapped_column(default=False, nullable=False)
    language: Mapped[Optional[str]] = mapped_column(
        db.String(50), nullable=True
    )
    imperial_units: Mapped[bool] = mapped_column(default=False, nullable=False)
    is_active: Mapped[bool] = mapped_column(default=False, nullable=False)
    email_to_confirm: Mapped[Optional[str]] = mapped_column(
        db.String(255), nullable=True
    )
    confirmation_token: Mapped[Optional[str]] = mapped_column(
        db.String(255), nullable=True
    )
    display_ascent: Mapped[bool] = mapped_column(default=True, nullable=False)
    accepted_policy_date: Mapped[Optional[datetime]] = mapped_column(
        TZDateTime, nullable=True
    )
    start_elevation_at_zero: Mapped[bool] = mapped_column(
        default=True, nullable=False
    )
    use_raw_gpx_speed: Mapped[bool] = mapped_column(
        default=False, nullable=False
    )
    use_dark_mode: Mapped[Optional[bool]] = mapped_column(
        default=False, nullable=True
    )
    manually_approves_followers: Mapped[bool] = mapped_column(
        default=True, nullable=False
    )
    hide_profile_in_users_directory: Mapped[bool] = mapped_column(
        default=True, nullable=False
    )
    workouts_visibility: Mapped[VisibilityLevel] = mapped_column(
        Enum(VisibilityLevel, name="visibility_levels"),
        server_default="PRIVATE",
        nullable=False,
    )
    map_visibility: Mapped[VisibilityLevel] = mapped_column(
        Enum(VisibilityLevel, name="visibility_levels"),
        server_default="PRIVATE",
        nullable=False,
    )
    suspended_at: Mapped[Optional[datetime]] = mapped_column(
        TZDateTime, nullable=True
    )
    role: Mapped[int] = mapped_column(
        CheckConstraint(
            f"role IN ({', '.join(UserRole.db_values())})",
            name="ck_users_role",
        ),
        nullable=False,
        default=UserRole.USER.value,
    )
    analysis_visibility: Mapped[VisibilityLevel] = mapped_column(
        Enum(VisibilityLevel, name="visibility_levels"),
        server_default="PRIVATE",
        nullable=False,
    )
    notification_preferences: Mapped[Optional[Dict]] = mapped_column(
        postgresql.JSONB, nullable=True
    )
    hr_visibility: Mapped[VisibilityLevel] = mapped_column(
        Enum(VisibilityLevel, name="visibility_levels"),
        server_default="PRIVATE",
        nullable=False,
    )
    segments_creation_event: Mapped[str] = mapped_column(
        Enum(*SEGMENTS_CREATION_EVENTS, name="segments_creation_events"),
        server_default="only_manual",
    )
    split_workout_charts: Mapped[bool] = mapped_column(
        server_default="true", nullable=False
    )
    messages_preferences: Mapped[Optional[Dict]] = mapped_column(
        postgresql.JSONB, nullable=True
    )
    missing_elevations_processing: Mapped[ElevationDataSource] = mapped_column(
        Enum(ElevationDataSource, name="elevation_data_source"),
        server_default="FILE",
        nullable=False,
    )
    calories_visibility: Mapped[VisibilityLevel] = mapped_column(
        Enum(VisibilityLevel, name="visibility_levels"),
        server_default="PRIVATE",
        nullable=False,
    )

    workouts: Mapped[List["Workout"]] = relationship(
        "Workout", lazy=True, back_populates="user"
    )
    records: Mapped[List["Record"]] = relationship(
        "Record", lazy=True, back_populates="user"
    )
    equipments: Mapped[List["Equipment"]] = relationship(
        "Equipment", lazy="select", back_populates="user"
    )
    received_follow_requests = relationship(
        FollowRequest,
        back_populates="to_user",
        primaryjoin=id == FollowRequest.followed_user_id,
        lazy="dynamic",
        cascade="all, delete-orphan",
    )
    sent_follow_requests = relationship(
        FollowRequest,
        back_populates="from_user",
        primaryjoin=id == FollowRequest.follower_user_id,
        lazy="dynamic",
        cascade="all, delete-orphan",
    )
    followers = relationship(
        "User",
        secondary="follow_requests",
        primaryjoin=and_(
            id == FollowRequest.followed_user_id,
            FollowRequest.is_approved == True,  # noqa
        ),
        secondaryjoin=and_(
            id == FollowRequest.follower_user_id,
            suspended_at == None,  # noqa
        ),
        lazy="dynamic",
        viewonly=True,
    )
    following = relationship(
        "User",
        secondary="follow_requests",
        primaryjoin=and_(
            id == FollowRequest.follower_user_id,
            FollowRequest.is_approved == True,  # noqa
        ),
        secondaryjoin=and_(
            id == FollowRequest.followed_user_id,
            suspended_at == None,  # noqa
        ),
        lazy="dynamic",
        viewonly=True,
    )
    comments: Mapped[List["Comment"]] = relationship(
        "Comment",
        lazy=True,
        back_populates="user",
        cascade="all, delete-orphan",
    )
    blocked_users = relationship(
        "BlockedUser",
        primaryjoin=id == BlockedUser.by_user_id,
        lazy="dynamic",
        viewonly=True,
    )
    blocked_by_users = relationship(
        "BlockedUser",
        primaryjoin=id == BlockedUser.user_id,
        lazy="dynamic",
        viewonly=True,
    )

    def __repr__(self) -> str:
        return f"<User {self.username!r}>"

    def __init__(
        self,
        username: str,
        email: str,
        password: str,
        created_at: Optional[datetime] = None,
    ) -> None:
        self.username = username
        self.email = email
        self.password = bcrypt.generate_password_hash(
            password, current_app.config.get("BCRYPT_LOG_ROUNDS")
        ).decode()
        self.created_at = (
            datetime.now(timezone.utc) if created_at is None else created_at
        )

    @staticmethod
    def encode_auth_token(user_id: int) -> str:
        """
        Generates the auth token
        :param user_id: -
        :return: JWToken
        """
        return get_user_token(user_id)

    @staticmethod
    def encode_password_reset_token(user_id: int) -> str:
        """
        Generates the auth token
        :param user_id: -
        :return: JWToken
        """
        return get_user_token(user_id, password_reset=True)

    @staticmethod
    def decode_auth_token(auth_token: str) -> Union[int, str]:
        """
        Decodes the auth token
        :param auth_token: -
        :return: integer|string
        """
        try:
            resp = decode_user_token(auth_token)
            is_blacklisted = BlacklistedToken.check(auth_token)
            if is_blacklisted:
                return "blacklisted token, please log in again"
            return resp
        except jwt.ExpiredSignatureError:
            return "signature expired, please log in again"
        except jwt.InvalidTokenError:
            return "invalid token, please log in again"

    def check_password(self, password: str) -> bool:
        return bcrypt.check_password_hash(self.password, password)

    @staticmethod
    def generate_password_hash(new_password: str) -> str:
        return bcrypt.generate_password_hash(
            new_password, current_app.config.get("BCRYPT_LOG_ROUNDS")
        ).decode()

    def get_user_id(self) -> int:
        return self.id

    @property
    def has_moderator_rights(self) -> bool:
        return has_moderator_rights(UserRole(self.role))

    @property
    def has_admin_rights(self) -> bool:
        return has_admin_rights(UserRole(self.role))

    @hybrid_property
    def workouts_count(self) -> int:
        return Workout.query.filter(Workout.user_id == self.id).count()

    @workouts_count.expression  # type: ignore
    def workouts_count(self) -> int:
        return (
            select(func.count(Workout.id))
            .where(Workout.user_id == self.id)
            .label("workouts_count")
        )

    @property
    def pending_follow_requests(self) -> List[FollowRequest]:
        return self.received_follow_requests.filter_by(updated_at=None).all()

    def send_follow_request_to(self, target: "User") -> FollowRequest:
        existing_follow_request = FollowRequest.query.filter_by(
            follower_user_id=self.id, followed_user_id=target.id
        ).first()
        if existing_follow_request:
            if existing_follow_request.is_rejected():
                raise FollowRequestAlreadyRejectedError()
            return existing_follow_request

        follow_request = FollowRequest(
            follower_user_id=self.id, followed_user_id=target.id
        )
        db.session.add(follow_request)
        if not target.manually_approves_followers:
            follow_request.is_approved = True
            follow_request.updated_at = datetime.now(timezone.utc)
        db.session.commit()

        return follow_request

    def unfollows(self, target: "User") -> None:
        existing_follow_request = FollowRequest.query.filter_by(
            follower_user_id=self.id, followed_user_id=target.id
        ).first()
        if not existing_follow_request:
            raise NotExistingFollowRequestError()

        db.session.delete(existing_follow_request)
        db.session.commit()
        return None

    def undoes_follow(self, follower: "User") -> None:
        existing_follow_request = FollowRequest.query.filter_by(
            followed_user_id=self.id, follower_user_id=follower.id
        ).first()
        if not existing_follow_request:
            raise NotExistingFollowRequestError()
        db.session.delete(existing_follow_request)
        db.session.commit()
        return None

    def _processes_follow_request_from(
        self, user: "User", approved: bool
    ) -> FollowRequest:
        follow_request = FollowRequest.query.filter_by(
            follower_user_id=user.id, followed_user_id=self.id
        ).first()
        if not follow_request:
            raise NotExistingFollowRequestError()
        if follow_request.updated_at is not None:
            raise FollowRequestAlreadyProcessedError()
        follow_request.is_approved = approved
        follow_request.updated_at = datetime.now(timezone.utc)
        db.session.commit()
        return follow_request

    def approves_follow_request_from(self, user: "User") -> FollowRequest:
        follow_request = self._processes_follow_request_from(
            user=user, approved=True
        )
        return follow_request

    def rejects_follow_request_from(self, user: "User") -> FollowRequest:
        follow_request = self._processes_follow_request_from(
            user=user, approved=False
        )
        return follow_request

    @staticmethod
    def follow_request_status(follow_request: Optional[FollowRequest]) -> str:
        if follow_request is None:
            return "false"
        if follow_request.is_approved:
            return "true"
        return "pending"

    def is_followed_by(self, user: "User") -> str:
        follow_request = FollowRequest.query.filter_by(
            follower_user_id=user.id, followed_user_id=self.id
        ).first()
        return self.follow_request_status(follow_request)

    def follows(self, user: "User") -> str:
        follow_request = FollowRequest.query.filter_by(
            follower_user_id=self.id, followed_user_id=user.id
        ).first()
        return self.follow_request_status(follow_request)

    def get_following_user_ids(self) -> List:
        return [following.id for following in self.following]

    def get_followers_user_ids(self) -> List:
        return [followers.id for followers in self.followers]

    def get_user_url(self) -> str:
        """Return user url on user interface"""
        return f"{current_app.config['UI_URL']}/users/{self.username}"

    def linkify_mention(self) -> str:
        return USER_LINK_TEMPLATE.format(
            profile_url=self.get_user_url(), username=f"@{self.username}"
        )

    def blocks_user(self, user: "User") -> None:
        if self.id == user.id:
            raise BlockUserException()

        db.session.execute(
            postgresql.insert(BlockedUser)
            .values(
                user_id=user.id,
                by_user_id=self.id,
                created_at=datetime.now(timezone.utc),
            )
            .on_conflict_do_nothing()
        )
        follow_request = FollowRequest.query.filter_by(
            follower_user_id=user.id,
            followed_user_id=self.id,
        ).first()
        if follow_request:
            db.session.delete(follow_request)
        db.session.commit()

    def unblocks_user(self, user: "User") -> None:
        BlockedUser.query.filter_by(
            user_id=user.id, by_user_id=self.id
        ).delete()
        db.session.commit()

    def is_blocked_by(self, user: "User") -> bool:
        return (
            BlockedUser.query.filter_by(
                user_id=self.id, by_user_id=user.id
            ).first()
            is not None
        )

    def get_blocked_user_ids(self) -> List:
        return [
            blocked_user.user_id for blocked_user in self.blocked_users.all()
        ]

    def get_blocked_by_user_ids(self) -> List:
        return [
            blocked_user.by_user_id
            for blocked_user in self.blocked_by_users.all()
        ]

    @property
    def suspension_action(self) -> Optional["ReportAction"]:
        if self.suspended_at is None:
            return None

        from fittrackee.reports.models import ReportAction

        return (
            ReportAction.query.filter(
                ReportAction.user_id == self.id,
                ReportAction.action_type == "user_suspension",
            )
            .order_by(ReportAction.created_at.desc())
            .first()
        )

    @property
    def sanctions_count(self) -> int:
        from fittrackee.reports.models import ReportAction

        return (
            ReportAction.query.filter(
                ReportAction.user_id == self.id,
                ReportAction.action_type.not_in(
                    [
                        "comment_unsuspension",
                        "user_unsuspension",
                        "user_warning_lifting",
                        "workout_unsuspension",
                    ]
                ),
            )
            .order_by(ReportAction.created_at.desc())
            .count()
        )

    @property
    def all_reports_count(self) -> Dict[str, int]:
        query = """
        SELECT (
                SELECT COUNT(*) AS created_reports_count
                FROM reports
                WHERE reports.reported_by = :user_id
            ),
            (
                SELECT COUNT(*) AS reported_count
                FROM reports
                WHERE reports.reported_user_id = :user_id
            ),
            (
                SELECT COUNT(*) AS sanctions_count
                FROM report_actions
                WHERE report_actions.user_id = :user_id
                  AND report_actions.action_type NOT IN (
                     'comment_unsuspension',
                     'user_unsuspension',
                     'user_warning_lifting',
                     'workout_unsuspension'
                )
        );"""
        result = db.session.execute(text(query), {"user_id": self.id}).one()
        return {
            "created_reports_count": result[0],
            "reported_count": result[1],
            "sanctions_count": result[2],
        }

    def update_notification_preferences(
        self, updated_preferences: Dict
    ) -> None:
        notification_preferences = {
            **(
                self.notification_preferences
                if self.notification_preferences
                else {}
            ),
            **updated_preferences,
        }
        validate(
            instance=notification_preferences,
            schema=NOTIFICATIONS_PREFERENCES_SCHEMA,
        )
        self.notification_preferences = notification_preferences
        db.session.commit()

    def update_message_preferences(self, updated_preferences: Dict) -> None:
        messages_preferences = {
            **(self.messages_preferences if self.messages_preferences else {}),
            **updated_preferences,
        }
        validate(
            instance=messages_preferences,
            schema=MESSAGE_PREFERENCES_SCHEMA,
        )
        self.messages_preferences = messages_preferences
        db.session.commit()

    def is_notification_enabled(self, notification_type: str) -> bool:
        if not self.notification_preferences:
            return True
        return self.notification_preferences.get(notification_type, True)

    def _get_records(self) -> List[Dict]:
        records = []
        sports_displayed_data = {}
        for record in self.records:
            if record.sport.id not in sports_displayed_data:
                sports_displayed_data[record.sport.id] = (
                    get_sport_displayed_data(record.sport, self)
                )
            sport_displayed_data = sports_displayed_data[record.sport.id]

            if (
                record.record_type == "HA"
                and not sport_displayed_data.display_elevation
            ):
                continue
            if (
                record.record_type in ["AP", "BP"]
                and not sport_displayed_data.display_pace
            ):
                continue
            if (
                record.record_type in ["AS", "MS"]
                and not sport_displayed_data.display_speed
            ):
                continue
            records.append(record.serialize())
        return records

    @property
    def calculated_missing_elevations_processing(
        self,
    ) -> "ElevationDataSource":
        if (
            self.missing_elevations_processing
            in [
                ElevationDataSource.OPEN_ELEVATION,
                ElevationDataSource.OPEN_ELEVATION_SMOOTH,
            ]
            and not current_app.config["OPEN_ELEVATION_API_URL"]
        ):
            return ElevationDataSource.FILE

        if (
            self.missing_elevations_processing == ElevationDataSource.VALHALLA
            and not current_app.config["VALHALLA_API_URL"]
        ):
            return ElevationDataSource.FILE

        return self.missing_elevations_processing

    def serialize(
        self,
        *,
        current_user: Optional["User"] = None,
        light: bool = True,
    ) -> Dict:
        if current_user is None:
            role = None
        else:
            role = (
                UserRole.AUTH_USER
                if current_user.id == self.id
                else UserRole(current_user.role)
            )

        serialized_user: Dict = {
            "created_at": self.created_at,
            "followers": self.followers.count(),
            "following": self.following.count(),
            "nb_workouts": self.workouts_count,
            "picture": self.picture is not None,
            "role": UserRole(self.role).name.lower(),
            "suspended_at": self.suspended_at,
            "username": self.username,
        }
        if is_auth_user(role) or has_moderator_rights(role):
            serialized_user["is_active"] = self.is_active
            serialized_user["email"] = self.email
        if (
            has_moderator_rights(role)
            and self.suspended_at
            and self.suspension_action
        ):
            serialized_user["suspension_report_id"] = (
                self.suspension_action.report_id
            )

        if current_user is not None and not is_auth_user(role):
            serialized_user["follows"] = self.follows(current_user)
            serialized_user["is_followed_by"] = self.is_followed_by(
                current_user
            )
            serialized_user["blocked"] = self.is_blocked_by(current_user)

        if light or not role:
            return serialized_user

        sports = []
        if self.workouts_count > 0:  # type: ignore
            sports = (
                db.session.query(Workout.sport_id)
                .filter(Workout.user_id == self.id)
                .group_by(Workout.sport_id)
                .order_by(Workout.sport_id)
                .all()
            )

        serialized_user = {
            **serialized_user,
            "bio": self.bio,
            "birth_date": self.birth_date,
            "first_name": self.first_name,
            "last_name": self.last_name,
            "location": self.location,
        }

        if role is not None:
            total = (0, "0:00:00", 0)
            if self.workouts_count > 0:  # type: ignore
                total = tuple(
                    db.session.query(
                        func.sum(Workout.distance),
                        func.sum(Workout.moving),
                        func.sum(Workout.ascent),
                    )
                    .filter(Workout.user_id == self.id)
                    .one()
                )

            serialized_user["nb_sports"] = len(sports)
            serialized_user["records"] = self._get_records()
            serialized_user["sports_list"] = [
                sport for sportslist in sports for sport in sportslist
            ]
            serialized_user["total_ascent"] = (
                float(total[2]) if total[2] else 0.0
            )
            serialized_user["total_distance"] = float(total[0])
            serialized_user["total_duration"] = str(total[1])

        if is_auth_user(role) or has_admin_rights(role):
            serialized_user["email_to_confirm"] = self.email_to_confirm

        if current_user and has_moderator_rights(UserRole(current_user.role)):
            reports_count = self.all_reports_count
            serialized_user["created_reports_count"] = reports_count[
                "created_reports_count"
            ]
            serialized_user["reported_count"] = reports_count["reported_count"]
            serialized_user["sanctions_count"] = reports_count[
                "sanctions_count"
            ]

        if is_auth_user(role):
            accepted_privacy_policy = None
            if self.accepted_policy_date:
                accepted_privacy_policy = (
                    current_app.config["privacy_policy_date"]
                    < self.accepted_policy_date
                )
            serialized_user = {
                **serialized_user,
                "accepted_privacy_policy": accepted_privacy_policy,
                "date_format": self.date_format,
                "display_ascent": self.display_ascent,
                "imperial_units": self.imperial_units,
                "language": self.language,
                "start_elevation_at_zero": self.start_elevation_at_zero,
                "timezone": self.timezone,
                "use_dark_mode": self.use_dark_mode,
                "use_raw_gpx_speed": self.use_raw_gpx_speed,
                "weekm": self.weekm,
                "map_visibility": self.map_visibility.value,
                "analysis_visibility": self.analysis_visibility.value,
                "workouts_visibility": self.workouts_visibility.value,
                "hr_visibility": self.hr_visibility.value,
                "segments_creation_event": self.segments_creation_event,
                "manually_approves_followers": (
                    self.manually_approves_followers
                ),
                "hide_profile_in_users_directory": (
                    self.hide_profile_in_users_directory
                ),
                "sanctions_count": self.sanctions_count,
                "notification_preferences": (
                    self.notification_preferences
                    if self.notification_preferences
                    else {}
                ),
                "split_workout_charts": self.split_workout_charts,
                "messages_preferences": (
                    self.messages_preferences
                    if self.messages_preferences
                    else {}
                ),
                "missing_elevations_processing": (
                    self.calculated_missing_elevations_processing
                ),
                "calories_visibility": self.calories_visibility.value,
            }

        return serialized_user


UserSportPreferenceEquipment = db.Table(
    "users_sports_preferences_equipments",
    db.Column(
        "user_id",
        db.Integer,
        db.ForeignKey("users.id", ondelete="CASCADE"),
        primary_key=True,
    ),
    db.Column(
        "sport_id",
        db.Integer,
        db.ForeignKey("sports.id", ondelete="CASCADE"),
        primary_key=True,
    ),
    db.Column(
        "equipment_id",
        db.Integer,
        db.ForeignKey("equipments.id", ondelete="CASCADE"),
        primary_key=True,
    ),
)


class UserSportPreference(BaseModel):
    __tablename__ = "users_sports_preferences"

    user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id"),
        primary_key=True,
    )
    sport_id: Mapped[int] = mapped_column(
        db.ForeignKey("sports.id"),
        primary_key=True,
    )
    color: Mapped[Optional[str]] = mapped_column(db.String(50), nullable=True)
    is_active: Mapped[bool] = mapped_column(default=True, nullable=False)
    stopped_speed_threshold: Mapped[float] = mapped_column(
        default=1.0, nullable=False
    )
    pace_speed_display: Mapped[PaceSpeedDisplay] = mapped_column(
        Enum(PaceSpeedDisplay, name="pace_speed_display"),
        server_default="SPEED",
        nullable=False,
    )

    default_equipments = relationship(
        "Equipment",
        secondary=UserSportPreferenceEquipment,
        primaryjoin=and_(
            user_id == UserSportPreferenceEquipment.c.user_id,
            sport_id == UserSportPreferenceEquipment.c.sport_id,
        ),
        lazy="dynamic",
        viewonly=True,
        back_populates="default_for_sports",
    )
    user: Mapped["User"] = relationship("User", lazy=True)

    def __init__(
        self,
        user_id: int,
        sport_id: int,
        stopped_speed_threshold: float,
        pace_speed_display: "PaceSpeedDisplay" = PaceSpeedDisplay.SPEED,
    ) -> None:
        self.user_id = user_id
        self.sport_id = sport_id
        self.is_active = True
        self.stopped_speed_threshold = stopped_speed_threshold
        self.pace_speed_display = pace_speed_display

    def serialize(self) -> Dict:
        return {
            "user_id": self.user_id,
            "sport_id": self.sport_id,
            "color": self.color,
            "is_active": self.is_active,
            "pace_speed_display": self.pace_speed_display,
            "stopped_speed_threshold": self.stopped_speed_threshold,
            "default_equipments": [
                equipment.serialize(current_user=self.user)
                for equipment in self.default_equipments.all()
            ],
        }


class BlacklistedToken(BaseModel):
    __tablename__ = "blacklisted_tokens"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    token: Mapped[str] = mapped_column(
        db.String(500), unique=True, nullable=False
    )
    expired_at: Mapped[int] = mapped_column(nullable=False)
    blacklisted_on: Mapped[datetime] = mapped_column(
        TZDateTime, nullable=False
    )

    def __init__(
        self, token: str, blacklisted_on: Optional[datetime] = None
    ) -> None:
        payload = jwt.decode(
            token,
            current_app.config["SECRET_KEY"],
            algorithms=["HS256"],
        )
        self.token = token
        self.expired_at = payload["exp"]
        self.blacklisted_on = (
            blacklisted_on if blacklisted_on else datetime.now(timezone.utc)
        )

    @classmethod
    def check(cls, auth_token: str) -> bool:
        return cls.query.filter_by(token=str(auth_token)).first() is not None


class UserTask(BaseModel):
    __tablename__ = "user_tasks"

    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    uuid: Mapped[UUID] = mapped_column(
        postgresql.UUID(as_uuid=True),
        default=uuid4,
        unique=True,
        nullable=False,
    )
    user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id", ondelete="CASCADE"), index=True
    )
    created_at: Mapped[datetime] = mapped_column(
        TZDateTime, nullable=False, default=aware_utc_now
    )
    updated_at: Mapped[Optional[datetime]] = mapped_column(
        TZDateTime, nullable=True, onupdate=aware_utc_now
    )
    task_type: Mapped[str] = mapped_column(
        Enum(*TASK_TYPES, name="task_types"), index=True
    )
    progress: Mapped[int] = mapped_column(nullable=False, default=0)
    errored: Mapped[bool] = mapped_column(nullable=False, default=False)
    aborted: Mapped[bool] = mapped_column(nullable=False, default=False)
    # can be input or output file
    file_size: Mapped[Optional[int]] = mapped_column(nullable=True)
    # relative or absolute path
    file_path: Mapped[Optional[str]] = mapped_column(
        db.String(255), nullable=True
    )
    errors: Mapped[Dict] = mapped_column(
        postgresql.JSONB, nullable=False, server_default="{}"
    )
    data: Mapped[Dict] = mapped_column(
        postgresql.JSONB, nullable=False, server_default="{}"
    )
    message_id: Mapped[Optional[str]] = mapped_column(
        db.String(36), nullable=True
    )

    def __init__(
        self,
        user_id: int,
        task_type: str,
        created_at: Optional[datetime] = None,
        data: Optional[Dict] = None,
        file_path: Optional[str] = None,
    ):
        self.user_id = user_id
        self.task_type = task_type
        self.created_at = (
            datetime.now(timezone.utc) if created_at is None else created_at
        )
        self.data = data if data else {}
        self.file_path = file_path
        if task_type == "workouts_archive_upload":
            self.errors = {
                "archive": None,
                "files": {},
            }

    @property
    def completed(self) -> bool:
        return self.progress == 100

    @property
    def short_id(self) -> str:
        return encode_uuid(self.uuid)

    def _get_user_data_export(
        self, serialized_task: Dict, *, for_admin: bool = False
    ) -> Dict:
        serialized_task = {
            **serialized_task,
            "file_size": (
                self.file_size
                if self.status == "successful" and self.file_size is not None
                else None
            ),
        }

        if for_admin:
            return serialized_task

        return {
            **serialized_task,
            "file_name": (
                self.file_path.split("/")[-1]
                if self.status == "successful" and self.file_path
                else None
            ),
        }

    @property
    def status(self) -> str:
        if self.errored:
            return "errored"
        elif self.aborted:
            return "aborted"
        elif self.progress == 0:
            return "queued"
        elif self.progress == 100:
            return "successful"
        return "in_progress"

    def _get_workouts_archive_upload(
        self, serialized_task: Dict, *, for_admin: bool = False
    ) -> Dict:
        total_files = len(self.data.get("files_to_process", []))

        serialized_task = {
            **serialized_task,
            "file_size": self.file_size,
            "files_count": total_files,
        }

        if for_admin:
            return serialized_task

        serialized_task = {
            **serialized_task,
            "sport_id": self.data.get("workouts_data", {}).get("sport_id"),
            "errored_files": self.errors,
            "new_workouts_count": self.data.get("new_workouts_count", 0),
            "progress": self.progress,
            "original_file_name": self.data.get("original_file_name"),
            "updated_at": self.updated_at,
        }
        return serialized_task

    def serialize(
        self,
        *,
        current_user: "User",
        for_admin: bool = False,
        task_user: Optional["User"] = None,
    ) -> Dict:
        if (
            current_user.id != self.user_id
            and not current_user.has_admin_rights
        ):
            raise UserTaskForbiddenException()

        if task_user and task_user.id != self.user_id:
            raise UserTaskException("invalid tasks user")

        serialized_task = {
            "id": self.short_id,
            "created_at": self.created_at,
            "status": self.status,
            "type": self.task_type,
        }

        if current_user.has_admin_rights and for_admin:
            serialized_task["message_id"] = self.message_id
            if task_user:
                serialized_task["user"] = task_user.serialize(
                    current_user=current_user, light=True
                )

        if self.task_type == "user_data_export":
            return self._get_user_data_export(
                serialized_task, for_admin=for_admin
            )

        return self._get_workouts_archive_upload(
            serialized_task, for_admin=for_admin
        )


@listens_for(UserTask, "after_delete")
def on_user_task_delete(
    mapper: Mapper, connection: Connection, old_record: "UserTask"
) -> None:
    @listens_for(db.Session, "after_flush", once=True)
    def receive_after_flush(session: Session, context: Any) -> None:
        Notification.query.filter(
            Notification.event_object_id == old_record.id,
        ).delete()
        if old_record.file_path:
            try:
                os.remove(get_absolute_file_path(old_record.file_path))
            except OSError:
                appLog.error("archive not found when deleting export request")


class Notification(BaseModel):
    __tablename__ = "notifications"
    __table_args__ = (
        db.UniqueConstraint(
            "from_user_id",
            "to_user_id",
            "event_type",
            "event_object_id",
            name="users_event_unique",
        ),
    )
    id: Mapped[int] = mapped_column(primary_key=True, autoincrement=True)
    uuid: Mapped[UUID] = mapped_column(
        postgresql.UUID(as_uuid=True),
        default=uuid4,
        unique=True,
        nullable=False,
    )
    from_user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    to_user_id: Mapped[int] = mapped_column(
        db.ForeignKey("users.id", ondelete="CASCADE"),
        index=True,
        nullable=False,
    )
    created_at: Mapped[datetime] = mapped_column(TZDateTime, nullable=False)
    marked_as_read: Mapped[bool] = mapped_column(
        db.Boolean, nullable=False, default=False
    )
    event_object_id: Mapped[Optional[int]] = mapped_column(nullable=True)
    event_type: Mapped[str] = mapped_column(db.String(50), nullable=False)

    def __init__(
        self,
        from_user_id: int,
        to_user_id: int,
        created_at: datetime,
        event_type: str,
        event_object_id: Optional[int] = None,
    ):
        if event_type not in NOTIFICATION_TYPES:
            raise InvalidNotificationTypeException()
        self.from_user_id = from_user_id
        self.to_user_id = to_user_id
        self.created_at = created_at
        self.event_type = event_type
        self.event_object_id = event_object_id

    @property
    def short_id(self) -> str:
        return encode_uuid(self.uuid)

    def serialize(self) -> Dict:
        serialized_notification = {
            "created_at": self.created_at,
            "id": self.short_id,
            "marked_as_read": self.marked_as_read,
            "type": self.event_type,
        }

        if self.event_type in ["follow", "follow_request"]:
            follow_request = FollowRequest.query.filter_by(
                follower_user_id=self.from_user_id,
                followed_user_id=self.to_user_id,
            ).one()
            from_user = follow_request.from_user
            to_user = follow_request.to_user
            return {
                **serialized_notification,
                "from": {
                    **from_user.serialize(),
                    "follows": from_user.follows(to_user),
                    "is_followed_by": from_user.is_followed_by(to_user),
                },
            }

        if self.event_type in [
            "comment_suspension",
            "comment_unsuspension",
            "user_warning",
            "user_warning_lifting",
            "workout_suspension",
            "workout_unsuspension",
        ]:
            from_user = None
        else:
            from_user = User.query.filter_by(id=self.from_user_id).first()
        to_user = User.query.filter_by(id=self.to_user_id).first()
        serialized_notification = {
            **serialized_notification,
            "from": (
                from_user.serialize(current_user=to_user)
                if from_user
                else None
            ),
        }

        if self.event_type == "workout_like":
            workout = Workout.query.filter_by(id=self.event_object_id).one()
            serialized_notification["workout"] = workout.serialize(
                user=to_user
            )

        if self.event_type in [
            "comment_like",
            "mention",
            "workout_comment",
        ]:
            comment = Comment.query.filter_by(id=self.event_object_id).one()
            serialized_notification["comment"] = comment.serialize(
                user=to_user
            )

        if self.event_type in [
            "report",
            "suspension_appeal",
            "user_warning_appeal",
        ]:
            from fittrackee.reports.models import Report, ReportActionAppeal

            if self.event_type in ["suspension_appeal", "user_warning_appeal"]:
                appeal = ReportActionAppeal.query.filter_by(
                    id=self.event_object_id
                ).one()
                report = Report.query.filter_by(
                    id=appeal.action.report_id
                ).one()
            else:
                report = Report.query.filter_by(id=self.event_object_id).one()
            serialized_notification["report"] = report.serialize(
                current_user=to_user
            )

        if self.event_type in [
            "comment_suspension",
            "comment_unsuspension",
            "user_warning",
            "user_warning_lifting",
            "workout_suspension",
            "workout_unsuspension",
        ]:
            from fittrackee.reports.models import Report, ReportAction

            report_action = ReportAction.query.filter_by(
                id=self.event_object_id
            ).one()
            serialized_notification["report_action"] = report_action.serialize(
                current_user=to_user
            )
            report = Report.query.filter_by(id=report_action.report_id).one()
            if report.object_type == "comment":
                comment = Comment.query.filter_by(
                    id=report.reported_comment_id
                ).first()
                serialized_notification["comment"] = (
                    comment.serialize(user=to_user) if comment else None
                )
            elif report.object_type == "workout":
                workout = Workout.query.filter_by(
                    id=report.reported_workout_id
                ).first()
                serialized_notification["workout"] = (
                    workout.serialize(user=to_user) if workout else None
                )

        if self.event_type == "workouts_archive_upload":
            task = UserTask.query.filter_by(id=self.event_object_id).first()
            if task:
                serialized_notification["task"] = {
                    "id": task.short_id,
                    "original_file_name": task.data.get("original_file_name"),
                }

        return serialized_notification



from typing import TYPE_CHECKING, Dict, List, Tuple, Union

from flask import Blueprint, current_app, request
from sqlalchemy import asc, desc, exc, nullslast

from fittrackee import db
from fittrackee.comments.exceptions import CommentForbiddenException
from fittrackee.oauth2.server import require_auth
from fittrackee.responses import (
    HttpResponse,
    InvalidPayloadErrorResponse,
    NotFoundErrorResponse,
    handle_error_and_return_response,
)
from fittrackee.users.exceptions import (
    UserAlreadyReactivatedException,
    UserAlreadySuspendedException,
    UserNotFoundException,
)
from fittrackee.users.models import User
from fittrackee.users.roles import UserRole
from fittrackee.utils import decode_short_id
from fittrackee.workouts.exceptions import WorkoutForbiddenException

from .exceptions import (
    InvalidReportActionException,
    InvalidReporterException,
    InvalidReportException,
    ReportNotFoundException,
    SuspendedObjectException,
    UserWarningExistsException,
)
from .models import (
    OBJECTS_ACTION_TYPES,
    REPORT_OBJECT_TYPES,
    Report,
    ReportActionAppeal,
)
from .reports_email_service import (
    ReportEmailService,
)
from .reports_service import ReportService

if TYPE_CHECKING:
    from sqlalchemy.sql.expression import UnaryExpression

reports_blueprint = Blueprint("reports", __name__)

REPORTS_PER_PAGE = 10
report_service = ReportService()


@reports_blueprint.route("/reports", methods=["POST"])
@require_auth(scopes=["reports:write"])
def create_report(auth_user: User) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Create a report.

    **Scope**: ``reports:write``

    **Example request**:

    .. sourcecode:: http

      POST /api/reports HTTP/1.1
      Content-Type: application/json

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 201 CREATED
      Content-Type: application/json

    :<json string note: note describing report
    :<json string object_id: id of content reported
    :<json string object_type: type of content reported (``comment``,
           ``workout`` or ``user``)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 201: report created
    :statuscode 400:
        - ``invalid payload``
        - ``users can not report their own comment``
        - ``users can not report their own profile``
        - ``users can not report their own workout``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
        - ``you do not have permissions, your account is suspended``
    :statuscode 404:
        - ``comment not found``
        - ``user not found``
        - ``workout not found``
    :statuscode 500: ``Error during comment save.``
    """
    data = request.get_json()
    object_id = data.get("object_id")
    object_type = data.get("object_type")
    note = data.get("note")
    if (
        not data
        or not note
        or not object_id
        or object_type not in REPORT_OBJECT_TYPES
    ):
        return InvalidPayloadErrorResponse()

    try:
        report_service.create_report(
            reporter=auth_user,
            note=note,
            object_id=object_id,
            object_type=object_type,
        )
        return {"status": "created"}, 201
    except (
        CommentForbiddenException,
        UserNotFoundException,
        WorkoutForbiddenException,
    ):
        return NotFoundErrorResponse(
            f"{object_type} not found "
            f"({'username' if object_type == 'user' else 'id'}: {object_id})"
        )
    except InvalidReporterException:
        return InvalidPayloadErrorResponse(
            "users can not report their own profile"
            if object_type == "user"
            else f"users can not report their own {object_type}s"
        )
    except (SuspendedObjectException, InvalidReportException) as e:
        return InvalidPayloadErrorResponse(str(e))
    except Exception as e:
        return handle_error_and_return_response(
            error=e,
            message="Error during report save.",
            status="fail",
            db=db,
        )


@reports_blueprint.route("/reports", methods=["GET"])
@require_auth(scopes=["reports:read"], role=UserRole.MODERATOR)
def get_reports(auth_user: User) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Get reports.

    **Scope**: ``reports:read``

    **Minimum role**: Moderator

    **Example requests**:

    - without parameters:

    .. sourcecode:: http

      GET /api/reports/ HTTP/1.1

    - with some query parameters:

    .. sourcecode:: http

      GET /api/reports?page=1&order=desc&order_by=created_at  HTTP/1.1

    **Example responses**:

    - returning at least one report:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

        {
          "pagination": {
            "has_next": false,
            "has_prev": false,
            "page": 1,
            "pages": 1,
            "total": 1
          },
          "reports": [
            {
              "created_at": "Sun, 01 Dec 2024 18:17:30 GMT",
              "id": 1,
              "is_reported_user_warned": false,
              "note": "<REPORT NOTE>",
              "object_type": "user",
              "reported_by": {
                "created_at": "Sun, 01 Dec 2024 17:27:56 GMT",
                "email": "moderator@example.com",
                "followers": 0,
                "following": 0,
                "is_active": true,
                "nb_workouts": 0,
                "picture": false,
                "role": "moderator",
                "suspended_at": null,
                "username": "moderator"
              },
              "reported_comment": null,
              "reported_user": {
                "blocked": false,
                "created_at": "Sun, 01 Dec 2024 17:27:49 GMT",
                "email": "sam@example.com",
                "followers": 0,
                "following": 0,
                "follows": "false",
                "is_active": true,
                "is_followed_by": "false",
                "nb_workouts": 1,
                "picture": false,
                "role": "user",
                "suspended_at": null,
                "username": "Sam"
              },
              "reported_workout": null,
              "resolved": false,
              "resolved_at": null,
              "resolved_by": null,
              "updated_at": null
            }
          ],
          "status": "success"
        }

    - returning no reports

    .. sourcecode:: http

      HTTP/1.1 200 OK
      Content-Type: application/json

        {
          "pagination": {
            "has_next": false,
            "has_prev": false,
            "page": 1,
            "pages": 0,
            "total": 0
          },
          "reports": [],
          "status": "success"
        }

    :query integer object_type: reported content type (``comment``, ``user``
           or ``workout``)
    :query string order: sorting order: ``asc``, ``desc`` (default: ``desc``)
    :query string order_by: sorting criteria: ``created_at`` or ``updated_at``
    :query integer page: page if using pagination (default: 1)
    :query boolean reporter: reporter username
    :query boolean resolved: filter on report status

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid payload``
        - ``invalid 'order_by'``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    """
    params = request.args.copy()
    page = int(params.get("page", 1))
    object_type = params.get("object_type")
    resolved = params.get("resolved", "").lower()
    column = params.get("order_by", "created_at")
    if column not in ["created_at", "updated_at"]:
        return InvalidPayloadErrorResponse("invalid 'order_by'")
    report_column = getattr(Report, column)
    order = params.get("order", "desc")
    order_clauses: List["UnaryExpression"] = [
        asc(report_column) if order == "asc" else desc(report_column)
    ]
    if column == "updated_at":
        order_clauses = [nullslast(order_clauses[0])]
    if column != "created_at":
        order_clauses.append(Report.created_at.desc())
    reporter_username = params.get("reporter", "")
    reporter = (
        User.query.filter(User.username == reporter_username).first()
        if reporter_username
        else None
    )

    filters = []
    if object_type:
        filters.append(Report.object_type == object_type)
    if resolved == "true":
        filters.append(Report.resolved == True)  # noqa
    elif resolved == "false":
        filters.append(Report.resolved == False)  # noqa
    if auth_user.role < UserRole.MODERATOR.value:
        filters.append(Report.reported_by == auth_user.id)
    elif reporter and reporter_username:
        filters.append(Report.reported_by == reporter.id)
    reports_pagination = (
        Report.query.filter(*filters)
        .order_by(*order_clauses)
        .paginate(page=page, per_page=REPORTS_PER_PAGE, error_out=False)
    )
    reports = reports_pagination.items
    return {
        "status": "success",
        "reports": [report.serialize(auth_user) for report in reports],
        "pagination": {
            "has_next": reports_pagination.has_next,
            "has_prev": reports_pagination.has_prev,
            "page": reports_pagination.page,
            "pages": reports_pagination.pages,
            "total": reports_pagination.total,
        },
    }, 200


@reports_blueprint.route("/reports/<int:report_id>", methods=["GET"])
@require_auth(scopes=["reports:read"], role=UserRole.MODERATOR)
def get_report(
    auth_user: User, report_id: int
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Get report.

    **Scope**: ``reports:read``

    **Minimum role**: Moderator

    **Example request**:

    .. sourcecode:: http

      GET /api/reports/1 HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

        {
          "report": {
            "comments": [],
            "created_at": "Sun, 01 Dec 2024 18:17:30 GMT",
            "id": 1,
            "is_reported_user_warned": false,
            "note": "<REPORT NOTE>",
            "object_type": "user",
            "report_actions": [],
            "reported_by": {
              "created_at": "Sun, 01 Dec 2024 17:27:56 GMT",
              "email": "moderator@example.com",
              "followers": 0,
              "following": 0,
              "is_active": true,
              "nb_workouts": 0,
              "picture": false,
              "role": "moderator",
              "suspended_at": null,
              "username": "moderator"
            },
            "reported_comment": null,
            "reported_user": {
              "blocked": false,
              "created_at": "Sun, 01 Dec 2024 17:27:49 GMT",
              "email": "sam@example.com",
              "followers": 0,
              "following": 0,
              "follows": "false",
              "is_active": true,
              "is_followed_by": "false",
              "nb_workouts": 1,
              "picture": false,
              "role": "user",
              "suspended_at": null,
              "username": "Sam"
            },
            "reported_workout": null,
            "resolved": false,
            "resolved_at": null,
            "resolved_by": null,
            "updated_at": null
          },
          "status": "success"
        }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 404:
        - ``report not found``
    """
    report = Report.query.filter_by(id=report_id).first()
    if not report or (
        not auth_user.has_moderator_rights
        and report.reported_by != auth_user.id
    ):
        return NotFoundErrorResponse(f"report not found (id: {report_id})")

    return {
        "status": "success",
        "report": report.serialize(auth_user, full=True),
    }, 200


@reports_blueprint.route("/reports/<int:report_id>", methods=["PATCH"])
@require_auth(scopes=["reports:write"], role=UserRole.MODERATOR)
def update_report(
    auth_user: User, report_id: int
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Update report.

    **Scope**: ``reports:write``

    **Minimum role**: Moderator

    **Example request**:

    .. sourcecode:: http

      PATCH /api/reports/1 HTTP/1.1

    **Example response** (report on user profile):

    .. sourcecode:: http

      HTTP/1.1 200 SUCCESS
      Content-Type: application/json

        {
          "report": {
            "comments": [
              {
                "comment": "<REPORT COMMENT>",
                "created_at": "Sun, 01 Dec 2024 18:21:38 GMT",
                "id": 1,
                "report_id": 1,
                "user": {
                  "created_at": "Sun, 01 Dec 2024 17:27:56 GMT",
                  "email": "moderator@example.com",
                  "followers": 0,
                  "following": 0,
                  "is_active": true,
                  "nb_workouts": 0,
                  "picture": false,
                  "role": "moderator",
                  "suspended_at": null,
                  "username": "moderator"
                }
              }
            ],
            "created_at": "Sun, 01 Dec 2024 18:17:30 GMT",
            "id": 1,
            "is_reported_user_warned": false,
            "note": "<REPORT NOTE>",
            "object_type": "user",
            "report_actions": [],
            "reported_by": {
              "created_at": "Sun, 01 Dec 2024 17:27:56 GMT",
              "email": "moderator@example.com",
              "followers": 0,
              "following": 0,
              "is_active": true,
              "nb_workouts": 0,
              "picture": false,
              "role": "moderator",
              "suspended_at": null,
              "username": "moderator"
            },
            "reported_comment": null,
            "reported_user": {
              "blocked": false,
              "created_at": "Sun, 01 Dec 2024 17:27:49 GMT",
              "email": "sam@example.com",
              "followers": 0,
              "following": 0,
              "follows": "false",
              "is_active": true,
              "is_followed_by": "false",
              "nb_workouts": 1,
              "picture": false,
              "role": "user",
              "suspended_at": null,
              "username": "Sam"
            },
            "reported_workout": null,
            "resolved": false,
            "resolved_at": null,
            "resolved_by": null,
            "updated_at": "Sun, 01 Dec 2024 18:21:38 GMT"
          },
          "status": "success"
        }

    :param string report_id: report id

    :<json string notes: report comment (mandatory)
    :<json boolean resolved: report status

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid payload``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
    :statuscode 404:
        - ``report not found``
    """
    data = request.get_json()
    comment = data.get("comment")
    if not data or not comment:
        return InvalidPayloadErrorResponse()

    try:
        report = report_service.update_report(
            report_id=report_id,
            moderator=auth_user,
            report_comment=comment,
            resolved=data.get("resolved"),
        )
    except ReportNotFoundException:
        return NotFoundErrorResponse(f"report not found (id: {report_id})")

    return {
        "status": "success",
        "report": report.serialize(auth_user, full=True),
    }, 200


@reports_blueprint.route("/reports/<int:report_id>/actions", methods=["POST"])
@require_auth(scopes=["reports:write"], role=UserRole.MODERATOR)
def create_action(
    auth_user: User, report_id: int
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Create report action.

    **Scope**: ``reports:write``

    **Minimum role**: Moderator

    **Example request**:

    .. sourcecode:: http

      POST /api/reports/1/actions HTTP/1.1

    **Example response** (report on user profile):

    .. sourcecode:: http

      HTTP/1.1 201 SUCCESS
      Content-Type: application/json

        {
          "report": {
            "comments": [
              {
                "comment": "<REPORT COMMENT>",
                "created_at": "Sun, 01 Dec 2024 18:21:38 GMT",
                "id": 1,
                "report_id": 1,
                "user": {
                  "created_at": "Sun, 01 Dec 2024 17:27:56 GMT",
                  "email": "moderator@example.com",
                  "followers": 0,
                  "following": 0,
                  "is_active": true,
                  "nb_workouts": 0,
                  "picture": false,
                  "role": "moderator",
                  "suspended_at": null,
                  "username": "moderator"
                }
              }
            ],
            "created_at": "Sun, 01 Dec 2024 18:17:30 GMT",
            "id": 1,
            "is_reported_user_warned": false,
            "note": "<REPORT NOTE>",
            "object_type": "user",
            "report_actions": [
              {
                "action_type": "user_warning",
                "appeal": null,
                "created_at": "Wed, 04 Dec 2024 09:12:25 GMT",
                "id": "Hv9KwVDtBHhyfvML7PHovq",
                "moderator": {
                  "created_at": "Sun, 01 Dec 2024 17:27:56 GMT",
                  "email": "moderator@example.com",
                  "followers": 0,
                  "following": 0,
                  "is_active": true,
                  "nb_workouts": 0,
                  "picture": false,
                  "role": "moderator",
                  "suspended_at": null,
                  "username": "moderator"
                },
                "reason": "<ACTION REASON>",
                "report_id": 1,
                "user": {
                  "blocked": false,
                  "created_at": "Sun, 01 Dec 2024 17:27:49 GMT",
                  "email": "sam@example.com",
                  "followers": 0,
                  "following": 0,
                  "follows": "false",
                  "is_active": true,
                  "is_followed_by": "false",
                  "nb_workouts": 1,
                  "picture": false,
                  "role": "user",
                  "suspended_at": null,
                  "username": "Sam"
                }
              }
            ],
            "reported_by": {
              "created_at": "Sun, 01 Dec 2024 17:27:56 GMT",
              "email": "moderator@example.com",
              "followers": 0,
              "following": 0,
              "is_active": true,
              "nb_workouts": 0,
              "picture": false,
              "role": "moderator",
              "suspended_at": null,
              "username": "moderator"
            },
            "reported_comment": null,
            "reported_user": {
              "blocked": false,
              "created_at": "Sun, 01 Dec 2024 17:27:49 GMT",
              "email": "sam@example.com",
              "followers": 0,
              "following": 0,
              "follows": "false",
              "is_active": true,
              "is_followed_by": "false",
              "nb_workouts": 1,
              "picture": false,
              "role": "user",
              "suspended_at": null,
              "username": "Sam"
            },
            "reported_workout": null,
            "resolved": false,
            "resolved_at": null,
            "resolved_by": null,
            "updated_at": "Sun, 01 Dec 2024 18:21:38 GMT"
          },
          "status": "success"
        }

    :param string report_id: report id

    :<json string action_type: action type (expected value:
           ``user_suspension``, ``user_unsuspension``, ``user_warning``,
           ``comment_suspension``, ``comment_unsuspension``,
           ``workout_suspension``, ``workout_unsuspension``)
    :<json string comment_id: id of comment affected by action (type:
           ``comment_suspension``, ``comment_suspension``)
    :<json string reason: text explaining the reason for the action
    :<json string username: username of user affected by action (type:
           ``user_suspension``, ``user_unsuspension``, ``user_warning``)
    :<json string workout_id: id of workout affected by action (type:
           ``workout_suspension``, ``workout_unsuspension``)

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid payload``
        - ``invalid 'action_type'``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
    :statuscode 404:
        - ``report not found``
    :statuscode 500:
        - ``Error during report save.``
    """
    data = request.get_json()
    action_type = data.get("action_type", "")
    reason = data.get("reason")
    if not data or not action_type:
        return InvalidPayloadErrorResponse()
    if (
        action_type == "user_warning_lifting"
        or action_type not in OBJECTS_ACTION_TYPES
    ):
        return InvalidPayloadErrorResponse("invalid 'action_type'")

    report = Report.query.filter_by(id=report_id).first()
    if not report:
        return NotFoundErrorResponse(f"report not found (id: {report_id})")

    try:
        action = report_service.create_report_action(
            report=report,
            moderator=auth_user,
            action_type=action_type,
            reason=reason,
            data=data,
        )
        db.session.flush()

        if current_app.config["CAN_SEND_EMAILS"]:
            report_action_email_service = ReportEmailService()
            report_action_email_service.send_report_action_email(
                report, action_type, reason, action
            )

        db.session.commit()

        return {
            "status": "success",
            "report": report.serialize(auth_user, full=True),
        }, 200
    except (
        InvalidReportActionException,
        UserAlreadyReactivatedException,
        UserAlreadySuspendedException,
        UserWarningExistsException,
    ) as e:
        return InvalidPayloadErrorResponse(str(e))
    except Exception as e:
        return handle_error_and_return_response(
            error=e,
            message="Error during report save.",
            status="fail",
            db=db,
        )


@reports_blueprint.route("/appeals/<string:appeal_id>", methods=["PATCH"])
@require_auth(scopes=["users:write"], role=UserRole.MODERATOR)
def process_appeal(
    auth_user: User, appeal_id: str
) -> Union[Dict, HttpResponse]:
    """
    Process appeal.

    **Scope**: ``users:write``

    **Minimum role**: Moderator

    **Example request**:

    .. sourcecode:: http

      POST /api/appeals/Z2Ze5qZrnMVmnDejPphASk HTTP/1.1

    **Example response** (report on user profile):

    .. sourcecode:: http

      HTTP/1.1 201 SUCCESS
      Content-Type: application/json

        {
          "appeal": {
            "approved": true,
            "created_at": "Wed, 04 Dec 2024 09:29:18 GMT",
            "id": "Z2Ze5qZrnMVmnDejPphASk",
            "moderator": {
              "created_at": "Sun, 01 Dec 2024 17:27:56 GMT",
              "email": "moderator@example.com",
              "followers": 0,
              "following": 0,
              "is_active": true,
              "nb_workouts": 0,
              "picture": false,
              "role": "moderator",
              "suspended_at": null,
              "username": "moderator"
            },
            "reason": "<REASON>",
            "text": "<APPEAL TEXT>",
            "updated_at": "Wed, 04 Dec 2024 09:30:21 GMT",
            "user": {
              "blocked": false,
              "created_at": "Sun, 01 Dec 2024 17:27:49 GMT",
              "email": "sam@example.com",
              "followers": 0,
              "following": 0,
              "follows": "false",
              "is_active": true,
              "is_followed_by": "false",
              "nb_workouts": 1,
              "picture": false,
              "role": "user",
              "suspended_at": null,
              "username": "Sam"
            }
          },
          "status": "success"
        }

    :param string appeal_id: appeal id

    :<json boolean approved: ``true`` if appeal is approved, ``false`` if
           rejected
    :<json string reason: text explaining why the appeal was approved or
           rejected

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 400:
        - ``invalid payload``
        - ``comment already reactivated``
        - ``user account has already been reactivated``
        - ``workout already reactivated``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
    :statuscode 404:
        - ``appeal not found``
    :statuscode 500:
        - ``Error during report save.``
    """
    appeal_uuid = decode_short_id(appeal_id)
    appeal = ReportActionAppeal.query.filter_by(uuid=appeal_uuid).first()

    if not appeal:
        return NotFoundErrorResponse(
            message=f"appeal not found (id: {appeal_id})"
        )

    data = request.get_json()
    reason = data.get("reason")
    if not data or "approved" not in data or not reason:
        return InvalidPayloadErrorResponse()

    try:
        new_report_action = report_service.process_appeal(
            appeal, auth_user, data
        )
        db.session.flush()

        if current_app.config["CAN_SEND_EMAILS"]:
            if new_report_action:
                report_action_email_service = ReportEmailService()
                report_action_email_service.send_report_action_email(
                    new_report_action.report,
                    new_report_action.action_type,
                    reason,
                    new_report_action,
                )
            if data["approved"] is False:
                action = appeal.action
                report_action_email_service = ReportEmailService()
                report_action_email_service.send_report_action_email(
                    action.report,
                    "appeal_rejected",
                    None,
                    action,
                )

        db.session.commit()
        return {
            "status": "success",
            "appeal": appeal.serialize(auth_user),
        }

    except InvalidReportActionException as e:
        return InvalidPayloadErrorResponse(str(e))
    except (exc.OperationalError, exc.IntegrityError, ValueError) as e:
        return handle_error_and_return_response(e, db=db)


@reports_blueprint.route("/reports/unresolved", methods=["GET"])
@require_auth(scopes=["reports:read"], role=UserRole.MODERATOR)
def get_unresolved_reports_status(
    auth_user: User,
) -> Union[Tuple[Dict, int], HttpResponse]:
    """
    Get if unresolved reports exist.

    **Scope**: ``reports:read``

    **Minimum role**: Moderator

    **Example request**:

    .. sourcecode:: http

      POST /api/reports/unresolved HTTP/1.1

    **Example response**:

    .. sourcecode:: http

      HTTP/1.1 201 SUCCESS
      Content-Type: application/json

        {
          "status": "success",
          "unresolved": true
        }

    :reqheader Authorization: OAuth 2.0 Bearer Token

    :statuscode 200: ``success``
    :statuscode 401:
        - ``provide a valid auth token``
        - ``signature expired, please log in again``
        - ``invalid token, please log in again``
    :statuscode 403:
        - ``you do not have permissions``
    """
    unresolved_reports = Report.query.filter(
        Report.resolved == False  # noqa
    ).count()
    return {
        "status": "success",
        "unresolved": unresolved_reports > 0,
    }, 200



from enum import Enum
from typing import List, Optional


class UserRole(Enum):
    OWNER = 100
    ADMIN = 50
    MODERATOR = 30
    AUTH_USER = 20  # only for app (not used in database)
    USER = 10

    @classmethod
    def values(cls) -> List[int]:
        return [e.value for e in cls]

    @classmethod
    def db_values(cls) -> List[str]:
        return [str(e.value) for e in cls if e.name != "AUTH_USER"]

    @classmethod
    def db_choices(cls) -> List[str]:
        return [e.name.lower() for e in cls if e.name != "AUTH_USER"]


def has_role_rights(role: Optional[UserRole], expected_role: UserRole) -> bool:
    return role is not None and role.value >= expected_role.value


def has_moderator_rights(role: Optional[UserRole]) -> bool:
    return has_role_rights(role, UserRole.MODERATOR)


def has_admin_rights(role: Optional[UserRole]) -> bool:
    return has_role_rights(role, UserRole.ADMIN)


def is_auth_user(role: Optional[UserRole]) -> bool:
    return role is not None and role == UserRole.AUTH_USER



"""add 'language' to 'User' table

Revision ID: f69f1e413bde
Revises: 27425324c9e3
Create Date: 2019-09-16 13:18:39.198777

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'f69f1e413bde'
down_revision = '27425324c9e3'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('users', sa.Column('language', sa.String(length=50), nullable=True))
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('users', 'language')
    # ### end Alembic commands ###



"""add stopped speed threshold to sports

Revision ID: 9842464bb885
Revises: cee0830497f8
Create Date: 2021-11-03 21:39:27.310371

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '9842464bb885'
down_revision = 'cee0830497f8'
branch_labels = None
depends_on = None


def upgrade():

    op.add_column(
        'sports',
        sa.Column('stopped_speed_threshold', sa.Float(), nullable=True),
    )

    op.execute(
        """
        UPDATE sports
        SET stopped_speed_threshold = 1
        WHERE label in (
              'Cycling (Sport)', 'Cycling (Transport)', 'Mountain Biking', 
              'Mountain Biking (Electric)', 'Rowing', 'Running',
              'Skiing (Alpine)'
        );
        UPDATE sports
        SET stopped_speed_threshold = 0.1
        WHERE label in (
              'Hiking', 'Skiing (Cross Country)', 'Trail', 'Walking'
        );
        """
    )
    op.alter_column('sports', 'stopped_speed_threshold', nullable=False)


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_column('sports', 'stopped_speed_threshold')
    # ### end Alembic commands ###



import os
from enum import Enum, IntEnum

TASKS_TIME_LIMIT = int(os.environ.get("TASKS_TIME_LIMIT", "1800")) * 1000


class TaskPriority(IntEnum):
    LOW = 100
    MEDIUM = 50
    HIGH = 0


class ElevationDataSource(str, Enum):  # to make enum serializable
    FILE = "file"
    OPEN_ELEVATION = "open_elevation"
    OPEN_ELEVATION_SMOOTH = "open_elevation_smooth"
    VALHALLA = "valhalla"


class PaceSpeedDisplay(str, Enum):
    PACE = "pace"  # min/km
    SPEED = "speed"
    PACE_AND_SPEED = "pace_and_speed"



from typing import TYPE_CHECKING, List, Optional

import feedgenerator
import mistune
from flask import current_app
from flask_babel import force_locale, lazy_gettext

from fittrackee.feeds.feeds.feed_item_template import FeedItemTemplate
from fittrackee.utils import clean_input
from fittrackee.visibility_levels import (
    can_view,
)

if TYPE_CHECKING:
    from fittrackee.users.models import User
    from fittrackee.workouts.models import Workout

KILOMETERS_TO_MILES_MULTIPLIER = 0.621371
METERS_TO_FEETS_MULTIPLIER = 3.280839895


class UserWorkoutsFeedService:
    def __init__(
        self,
        user: "User",
        workouts: List["Workout"],
        lang: str = "en",
        use_imperial_units: bool = False,
        with_description: bool = False,
    ):
        self.fittrackee_url = current_app.config["UI_URL"]
        self.user = user
        self.workouts = workouts
        self.lang = lang if lang in current_app.config["LANGUAGES"] else "en"
        self.with_description = with_description

        # to display distance and speed
        self.distance_multiplier = (
            KILOMETERS_TO_MILES_MULTIPLIER if use_imperial_units else 1.0
        )
        self.distance_unit = "mi" if use_imperial_units else "km"
        # to display elevation and ascent/descent
        self.elevation_multiplier = (
            METERS_TO_FEETS_MULTIPLIER if use_imperial_units else 1.0
        )
        self.elevation_unit = "ft" if use_imperial_units else "m"

        self.feed = self.init_feed()
        self.feed_template = FeedItemTemplate(
            current_app.config["FEEDS_TEMPLATES_FOLDER"],
            current_app.config["TRANSLATIONS_FOLDER"],
            current_app.config["LANGUAGES"],
        )

    def init_feed(self) -> "feedgenerator.Rss201rev2Feed":
        user_ui_url = f"{self.fittrackee_url}/users/{self.user.username}"
        with force_locale(self.lang):
            feed = feedgenerator.Rss201rev2Feed(
                description=(
                    lazy_gettext(
                        "Latest public workouts on FitTrackee from %(username)s",  # noqa
                        username=self.user.username,
                    )
                ),
                feed_url=f"{user_ui_url}/workouts.rss",
                language=self.lang,
                link=user_ui_url,
                title=(
                    lazy_gettext(
                        "%(username)s's workouts feed",
                        username=self.user.username,
                    )
                ),
            )
        return feed

    def _get_distance(self, value: Optional[float]) -> Optional[float]:
        if value is None:
            return None
        return round(float(value) * self.distance_multiplier, 3)

    def _get_elevation(
        self, value: Optional[float], can_view_altitude: bool = True
    ) -> Optional[float]:
        if value is None or not can_view_altitude:
            return None
        return round(float(value) * self.elevation_multiplier, 2)

    def _get_workout_item(
        self, fittrackee_url: str, workout: "Workout"
    ) -> dict:
        can_view_map = workout.map_id and can_view(
            workout,
            "calculated_map_visibility",
            user=None,
        )
        can_view_altitude = can_view(
            workout,
            "calculated_analysis_visibility",
            user=None,
        )
        data = {
            "workout_map": (
                f"{fittrackee_url}/api/workouts/map/{workout.map_id}"
                if can_view_map
                else None
            ),
            "workout_title": (
                clean_input(workout.title) if workout.title else ""
            ),
            "with_description": self.with_description and workout.description,
            "workout_distance": self._get_distance(workout.distance),
            "workout_distance_unit": self.distance_unit,
            "workout_min_altitude": self._get_elevation(
                workout.min_alt, can_view_altitude
            ),
            "workout_max_altitude": self._get_elevation(
                workout.max_alt, can_view_altitude
            ),
            "workout_ascent": self._get_elevation(workout.ascent),
            "workout_descent": self._get_elevation(workout.descent),
            "workout_elevation_unit": self.elevation_unit,
            "workout_duration": workout.moving,
            "sport_label": workout.sport.label,
        }
        return self.feed_template.get_item_data("workout", self.lang, data)

    def generate_user_workouts_feed(self) -> str:
        markdown = mistune.create_markdown(
            escape=False, plugins=["url", "speedup"]
        )
        for workout in self.workouts:
            item_data = self._get_workout_item(self.fittrackee_url, workout)
            workout_description = (
                clean_input(
                    markdown(workout.description),  # type: ignore[arg-type]
                    for_markdown_renderer=True,
                )
                if self.with_description and workout.description
                else ""
            )
            self.feed.add_item(
                title=item_data["title.txt"],
                link=f"{self.fittrackee_url}/workouts/{workout.short_id}",
                pubdate=workout.workout_date,
                description=item_data["body.html"] + workout_description,
            )
        return self.feed.writeString("UTF-8")



import json
from typing import TYPE_CHECKING, Dict, List, Optional

import geopandas as gpd
from shapely import Point
from sqlalchemy import func, select

from fittrackee import db
from fittrackee.utils import decode_short_id
from fittrackee.workouts.constants import (
    WGS84_CRS,
)
from fittrackee.workouts.exceptions import (
    InvalidCoordinatesException,
    InvalidRadiusException,
)
from fittrackee.workouts.models import WorkoutSegment
from fittrackee.workouts.utils.sports import get_sport_displayed_data

if TYPE_CHECKING:
    from fittrackee.users.models import User
    from fittrackee.workouts.models import Sport, Workout


def get_geojson_from_segments(
    workout: "Workout",
    *,
    segment_short_id: Optional[str] = None,
) -> Optional[Dict]:
    """
    To refactor when using segment uuid
    """
    filters = [WorkoutSegment.workout_id == workout.id]
    if segment_short_id is not None:
        segment_uuid = decode_short_id(segment_short_id)
        filters.append(WorkoutSegment.uuid == segment_uuid)
    geom_subquery = select(WorkoutSegment.geom).filter(*filters).subquery()
    subquery = (
        func.ST_Collect(geom_subquery.c.geom)
        if segment_short_id is None
        else geom_subquery.c.geom
    )
    geojson = db.session.scalar(func.ST_AsGeoJSON(subquery))
    if geojson:
        return json.loads(geojson)
    return None


def get_chart_data_from_segment_points(
    segments_points: List[List[Dict]],
    sport: "Sport",
    *,
    user: Optional["User"],
    workout_ave_cadence: Optional[int],
    can_see_heart_rate: bool,
) -> List:
    """
    Return data needed to generate chart with:
    - speed depending on sport and user sport preferences
    - pace depending on sport and user sport preferences
    - elevation (if available)
    - heart rate (if available)
    - cadence (if available)
    - power (if available)
    """
    chart_data = []
    sport_data_visibility = get_sport_displayed_data(sport, user)
    return_cadence = (
        workout_ave_cadence and sport_data_visibility.display_cadence
    )
    total_distance = 0

    for segment_points in segments_points:
        if not segment_points:
            continue

        points_count = len(segment_points)
        first_point = segment_points[0]
        first_point_duration = (
            first_point["duration"] if len(segments_points) == 1 else 0
        )

        for index, point in enumerate(segment_points, start=1):
            distance = round((point["distance"]) / 1000 + total_distance, 2)
            data = {
                "distance": distance,
                "duration": point["duration"] - first_point_duration,
                "latitude": point["latitude"],
                "longitude": point["longitude"],
                "time": point["time"],
            }
            if sport_data_visibility.display_elevation and point.get(
                "elevation"
            ):
                data["elevation"] = point["elevation"]
            if sport_data_visibility.display_pace and "pace" in point:
                data["pace"] = point["pace"]
            if "speed" in point and (
                sport_data_visibility.display_speed
                # in case pace has not yet been recalculated
                or (sport_data_visibility.display_pace and "pace" not in point)
            ):
                data["speed"] = point["speed"]
            if return_cadence and "cadence" in point:
                data["cadence"] = (
                    point["cadence"] * 2
                    if sport_data_visibility.display_spm_cadence
                    else point["cadence"]
                )
            if sport_data_visibility.display_power and "power" in point:
                data["power"] = point["power"]
            if can_see_heart_rate and "heart_rate" in point:
                data["hr"] = point["heart_rate"]

            if index == points_count:
                total_distance = distance
            chart_data.append(data)

    return chart_data


def get_buffered_location(coordinates: str, radius_str: str) -> str:
    """
    coordinates: latitude,longitude
    radius: distance in kilometers
    """
    try:
        radius = float(radius_str)
    except ValueError as e:
        raise InvalidRadiusException() from e
    if radius <= 0:
        raise InvalidRadiusException()

    try:
        latitude, longitude = coordinates.split(",")
        point = Point(float(longitude), float(latitude))
    except ValueError as e:
        raise InvalidCoordinatesException() from e

    gdf = gpd.GeoDataFrame(geometry=[point], crs=WGS84_CRS)
    gdf_in_meters = gdf.to_crs(gdf.estimate_utm_crs())
    buffered_gdf_in_meters = gdf_in_meters.buffer(radius * 1000)
    buffered_gdf = buffered_gdf_in_meters.to_crs(WGS84_CRS)
    return f"SRID={WGS84_CRS};{buffered_gdf.loc[0]}"
